/* Copyright (c) 1996 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     BasicPlanner

   PURPOSE
     provide basic bookkeeping for planning algo.

   NOTES
     
   AUTHOR
     David Hsu  
   
     dyhsu@cs.stanford.edu
     http://robotics.stanford.edu/~dyhsu 
                  
   HISTORY
     created July, 1996

***************************************************************************/


#include "basicPlanner.H"
#include "global.H"

void BasicPlanner::query(Config& ic, Config& gc, ostream& pfile)
{
    init = ic;
    goal = gc;
    query(pfile);
}


/* bgn
 *   read the input from the streams and make the planner ready
 *
 *   rsrc : stream for robot data
 *   osrc : stream for obstacle data
 *   asrc : stream for cspace data
 */

void BasicPlanner::bgn(istream& rsrc, istream& osrc, istream &asrc)
{
    // make robot object
    rsrc >> r;
    cerr << "Robot max radius " << r.radius() << endl;

    // make obstacle object
    osrc >> obs;

    // make c-space object
    asrc >> csp;
    cerr << "Axes: \n" << csp;
}


/* writePath
 *   write out the path if it is found
 */

void BasicPlanner::writePath(PtrConfig ic, PtrConfig gc, ostream& os)
{
    int nc = 0;
    int i, marker;              
    PtrConfig node;

    // count the number of nodes in the path 
    node = ic;
    while ( node ) {
        node = node->link; nc++;
    }

    node = gc;
    while ( node ) {
        node = node->link; nc++;
    }

    // allocate memory for the path
    PtrConfig* path = new PtrConfig[nc];

    // write the portion of the path in the initial component into path
    node = ic; i = 0;
    while ( node ) {
        path[i++] =  node;
        node = node->link;
    }
    marker = i - 1;

    // reverse the path
    for (i = 0; i < ( marker + 1) / 2; i++ ) 
	swap(path[i], path[marker-i]);


    // write the portion of the path in the goal component into path
    node = gc; i = marker + 1;
    while ( node ) {
        path[i++] =  node;
        node = node->link;
    }
    marker = i;                         /* mark the number of nodes */

    // link them into a list
    for (i = 0; i < marker - 1; i++) 
        path[i]->link = path[i+1];
    path[i++]->link = NULL;
    
    // smooth the path if required
    if ( Ct.sm ) smoothPath(path, marker);

    // write the path data file
    // if interpolation is required, subdivide the configurations
    if ( Ct.it ){
	ConfigArray* simplePath;

	for (i=0; i < 16; i++)
	    os << (*path[0]) << endl;
        for (node = path[0]; node->link != NULL; node = node->link) {
	    simplePath = interpolate(*node, *(node->link), r );
	    simplePath->printOmitLast(os);
	    delete simplePath;
        }

       // write out the last node
	for (i=0; i < 16; i++)
	    os <<  	(*node ) << endl;
    }

    // otherwise, just dump out the path
    else { 
        for (node = path[0]; node != NULL; node = node->link )
	    os <<  	(*node ) << endl;
        os << endl;
    } 

    // deallocate memory
    delete [] path; 
}


/* smoothPath
 *   shorten the path by cutting corners
 *   Running time is O(n log n)
 *
 *  path: array of pointers each of which points to a config in the path
 *   n  : the number of elements in  path
 */

void BasicPlanner::smoothPath(PtrConfig* path, int n)
{       
    int l=0, r0, r1, m;
    Robot* rob = &r;

    // l marks the beginning of current segment, the end of current segment
    // is in between r0 and r1
    // invariant : the line segment bewteen path[l] and path[r0] lies in
    // C-free

    while (l != n-1) {  /* keep going if we are not at the end of path */

        r0 = l+1; r1 = n-1; m = r1 ;
        while ( r1 > r0 + 1  ) {
            if ( rob->connectPair( *path[l], *path[m]) )
                r0 = m;
            else
                r1 = m;
            m = (r0 + r1) / 2 ;
        }
        
        // line path[l] with path[r0] and move onto the portion of the path
        path[l]->link = path[r0];
        l = r0;
    }
}


ostream& operator<< (ostream& os, BasicPlanner& p)
{
    os << "Init Configuration:\n\t" << p.init  << "\tdist = " <<
	p.init.dist << '\n'
	<< "Goal Configuration:\n\t" << p.goal  << "\tdist = " <<
	p.goal.dist << "\n\n";
    if (p.found)
	os << "Path found\n";
    else
	os << "Path not found\n";
    return os;
}

istream& operator>> (istream& is, BasicPlanner& p)
{
    is >> p.init >> p.goal;
    return is;
}

    
    


    
